#!/usr/bin/python3
# -*- coding: utf-8 -*-

#
# A module to access data from the DFG-funded SPP Computational Literary Studies
#
# Usage:
# from sppcls import sppcls
# df = sppcls.load_df("judenbuche", ["events", "keypassages"])
#
# Author: rja
#
# Changes:
# 2022-09-02 (rja)
# - disabled parsing of comments
# 2022-08-26 (rja)
# - fixed usage example
# - fixed column name (s/word/token/)
# 2022-08-25 (rja)
# - initial version

import pandas as pd
import asyncio

# default repository URL
SPP_CLS_GITLAB = "https://cls-gitlab.phil.uni-wuerzburg.de/spp-cls-data-exchange/spp-cls_annotationtables_data/-/raw/main/"


# FIXME: do we want to support more/all pandas.DataFrame kwargs?
def load_df(work, projects, repository=SPP_CLS_GITLAB, na_values=""):
    """Generates a dataframe with the data of the projects for work."""
    dataframes = [_download(work, p, repository, na_values) for p in projects]
    dataframe = _merge(dataframes, projects)
    _check(dataframe, dataframes)
    return dataframe


async def load_df_async(work, projects, repository=SPP_CLS_GITLAB, na_values=""):
    return await asyncio.get_running_loop().run_in_executor(None, load_df, work, projects, repository, na_values)


def _download(work, project, repository, na_values):
    # FIXME: add proper URL escaping
    url = repository + "/" + work + "/" + project + ".tsv"
    return pd.read_csv(url, sep='\t', na_values=na_values, comment='#')


def _merge(dataframes, projects):
    """Merge the dataframes using a JOIN."""
    dataframe = dataframes[0]
    for df in dataframes[1:]:
        dataframe = pd.merge(dataframe, df, on=["id", "token"], suffixes=_suffixes(projects))
    return dataframe


def _suffixes(projects, delim="_"):
    l = _shortest_prefix_length(projects)
    return [delim + p[:l] for p in projects]


def _shortest_prefix_length(projects):
    """Find the smallest l such that the first l chars of projects are uniqe."""
    # FIXME: avoid fixed length
    for i in range(1, 10):
        if len(set([p[:i] for p in projects])) == len(projects):
            return i
    return None


def _check(dataframe, dataframes):
    for df in dataframes:
        assert len(df) == len(dataframe)


if __name__ == '__main__':
    pass



import sys
from argparse import ArgumentParser
from sppcls.check import check

# Only import tokenise if spacy is found. It is overkill to always install spacy which is normally not needed as
# tokenization is only done a single time per literary work.
try:
    from sppcls.tokenise import tokenise
except ModuleNotFoundError:
    pass


def main(argv=None):
    argument_parser = ArgumentParser(description="Accessing and processing data from the DFG-funded SPP Computational Literary Studies")

    subparsers = argument_parser.add_subparsers(dest="command")
    subparsers.required = True

    parser_tokenise = subparsers.add_parser("tokenise", help="Tokenize text file and create output tsv.",
                                            description="Tokenize text file and create output tsv.")

    parser_tokenise.add_argument("input_file", help="Path to the input txt file.")
    parser_tokenise.add_argument("output_folder", help="Path to the output folder where the output tsv will be saved.")

    parser_check = subparsers.add_parser("check", help="Compare two tsv files and check that the structures matches.",
                                         description="Compare two tsv files and check that the structures matches.")

    parser_check.add_argument("org_tokens_file_path", metavar="org-tokens-file-path",
                                 help="Path to the original tokens tsv file")
    parser_check.add_argument("project_tokens_file_path", metavar="project-tokens-file-path",
                                 help="Path to the project tokens tsv file")

    args = argument_parser.parse_args(argv)

    if args.command == "tokenise":
        in_file_path = args.input_file
        output_folder_path = args.output_folder
        tokenise.main(in_file_path, output_folder_path)

    elif args.command == "check":
        org_tokens_file_path = args.org_tokens_file_path
        project_tokens_file_path = args.project_tokens_file_path
        check.main(org_tokens_file_path, project_tokens_file_path)


if __name__ == '__main__':
    sys.exit(main())



import csv
from os.path import join, splitext, basename
import spacy


def tokenise(input_text: str):
    nlp = spacy.load("de_core_news_lg")
    doc = nlp(input_text)
    result = []
    for sent_idx, sent in enumerate(doc.sents):
        result.append({"new_sentence": sent_idx})

        for token in sent:
            result.append((token.i, token.text.replace("\n", " "), token.idx, token.idx + len(token.text)))
    return result


def main(in_file_path, output_folder_path):
    in_file = open(in_file_path, "rt", encoding="utf-8")
    input_filename = splitext(basename(in_file_path))[0]
    input_text = in_file.read()
    tokens = tokenise(input_text)
    out_file = open(join(output_folder_path, f"{input_filename}_tokenized.tsv"), "wt", encoding="utf-8")
    writer = csv.writer(out_file, delimiter="\t", lineterminator="\n")
    writer.writerow(["id", "token", "start", "end"])
    for token in tokens:
        if 'new_sentence' in token:
            writer.writerow([f"# new sentence id {token['new_sentence']}"])
        else:
            writer.writerow(token)

    out_file.close()


if __name__ == '__main__':
    pass



import csv
from os.path import isfile


def read_tokens(token_file_path):
    result = []

    with open(token_file_path, 'r', encoding='utf-8') as token_file:
        reader = csv.reader(token_file, delimiter='\t')

        header = next(reader, None)
        column_count = len(header)

        for row in reader:
            if len(row) == 1:
                result.append(row[0])
            else:
                result.append(row)

    return column_count, result


def check_for_issues(org_tokens, project_tokens, column_count):

    if len(org_tokens) != len(project_tokens):
        raise Exception("Token lists differ in length")

    for line, (org_token, project_token) in enumerate(zip(org_tokens, project_tokens)):
        if isinstance(org_token, str):
            if not isinstance(project_token, str):
                raise Exception(f"Tokens don't match (line {line + 2})")

            continue

        if len(project_token) != column_count:
            raise Exception(f"Incorrect column count (line {line + 2})")

        for i in range(0, len(org_token)):
            if org_token[i] != project_token[i]:
                raise Exception(f"Tokens don't match (line {line + 2}, position {i + 1})")


def main(org_tokens_file_path, project_tokens_file_path):

    if not isfile(org_tokens_file_path) or not org_tokens_file_path.endswith(".tsv"):
        raise Exception("Invalid original tokens tsv file path")

    if not isfile(project_tokens_file_path) or not project_tokens_file_path.endswith(".tsv"):
        raise Exception("Invalid project tokens tsv file path")

    print(f'Compare {org_tokens_file_path} and {project_tokens_file_path}')

    _, org_tokens = read_tokens(org_tokens_file_path)
    project_column_count, project_tokens = read_tokens(project_tokens_file_path)

    check_for_issues(org_tokens, project_tokens, project_column_count)

    print("No issues found")


if __name__ == '__main__':
    pass



import argparse
import csv
from dataclasses import dataclass
from difflib import SequenceMatcher
import json
import os
from pathlib import Path
from typing import List
import zipfile

import requests
import spacy

from sppcls.tokenise import tokenise

spacy.tokens.Token.set_extension("event_anno_id", default=None)
spacy.tokens.Token.set_extension("event_kind", default=None)

DATASET_URL = "https://github.com/forTEXT/EvENT_Dataset/archive/refs/tags/v.1.1.zip"
CACHE_PATH = Path(".cache")
EVENT_ZIP_PATH = CACHE_PATH / "events.zip"

parser = argparse.ArgumentParser("Create event data from event ")
parser.add_argument("tsv_input_path")
parser.add_argument("tsv_output_path")
parser.add_argument("text_name", default="Die Judenbuche", nargs="?")
parser.add_argument("zip_path", default=EVENT_ZIP_PATH, nargs="?")


class TSVDialect(csv.excel_tab):
    # We had \r\n line endings but they disappeared in our process
    # likely some over eager text editor... So let's just do newlines
    lineterminator = "\n"


@dataclass
class UniversalAnnotation:
    id: int
    text: str
    start: int
    end: int

    @classmethod
    def from_row(cls, row):
        id, text, start, end = row
        return UniversalAnnotation(
            id,
            text,
            int(start),
            int(end),
        )

    @staticmethod
    def to_text(sentence: List["UniversalAnnotation"]):
        out = []
        for element in sentence:
            out.append(element.text)
        return "\t".join(out)


def load_dataset(zip_file: zipfile.ZipFile):
    data = json.load(zip_file.open("EvENT_Dataset-v.1.1/Annotations_EvENT.json"))
    text_dir = zipfile.Path(zip_file, "EvENT_Dataset-v.1.1/Plain_Texts/")
    for full_path in text_dir.iterdir():
        full_text = full_path.read_text()
        path = Path(full_path.filename)
        simple_name = path.name.split("_")[-1][:-4]
        for work_name in data.keys():
            if simple_name.lower() in work_name.lower():
                data[work_name]["full_text"] = full_text
    return data



def get_json_dataset(zip_path: Path):
    if os.path.isfile(zip_path):
        data = load_dataset(zipfile.ZipFile(str(zip_path)))
    else:
        os.makedirs(zip_path.parent, exist_ok=True)
        response = requests.get(DATASET_URL)
        out_file = open(zip_path, "wb")
        assert response.status_code == 200
        try:
            for chunk in response.iter_content(chunk_size=512 * 1024):
                if chunk is not None:
                    out_file.write(chunk)
        except Exception as err:
            os.unlink(zip_path)
            raise err
        out_file.close()
        zip_file = zipfile.ZipFile(str(zip_path))
        data = load_dataset(zip_file)
    return data


def add_annotation_to_tokens(doc, gold_standard):
    spans = []
    for i, annotation in enumerate(gold_standard):
        for span in annotation["spans"]:
            spans.append(((span[0], span[1]), (i, annotation["tag"])))
    # This is quadratic and could be heavily optmized but ¯\_(ツ)_/¯
    for token in doc:
        for (start, end), (anno_id, tag) in spans:
            if start <= token.idx and end >= (token.idx + len(token)):
                token._.event_anno_id = anno_id
                token._.event_kind = tag


def get_reference_token_annotations(op_codes, our_doc, strs_reference):
    annotations = [] # Triple: (was_annotated, annnotation_id, annotation_tag)
    for op, reference_start, reference_end, ours_start, ours_end in op_codes:
        if op == "insert":
            # We don't really care, our annotations are just not added
            pass
        elif op == "delete":
            # We need to set the output tokens to unseen/unannotated
            for i in range(reference_start, reference_end):
                annotations.append((False, "-", "-"))
        elif op == "replace":
            # We just transfer the annotation to the new tokens
            if len(set(t._.event_anno_id for t in our_doc[ours_start:ours_end])) == 1:
                for _ in range(reference_start, reference_end):
                    annotations.append((True, our_doc[ours_start]._.event_anno_id or "-", our_doc[ours_start]._.event_kind or "-"))
            elif (reference_end - reference_start > 5) and reference_end - reference_start == ours_end - ours_start:
                # They are the same length, we are okay up to 5 and with the same length
                for i in range(ours_start, ours_end):
                    annotations.append((True, our_doc[i]._.event_anno_id or "-", our_doc[i]._.event_kind or "-"))
            else:
                for _ in range(reference_start, reference_end):
                    annotations.append((False, "-", "-"))
        elif op == "equal":
            for i_ours, i_ref in zip(range(ours_start, ours_end), range(reference_start, reference_end)):
                assert our_doc[i_ours].text.strip() == strs_reference[i_ref].strip()
                annotations.append((True, our_doc[i_ours]._.event_anno_id or "-", our_doc[i_ours]._.event_kind or "-"))
    assert len(annotations) == len(strs_reference)
    return annotations


def main(input_json: Path, reference_tsv: Path, text_name: str, out_path: Path):
    nlp = spacy.load("de_core_news_lg")
    data = get_json_dataset(input_json)
    gold_standard_annos = data[text_name]["gold_standard"]
    reader = csv.reader(open(reference_tsv), dialect=TSVDialect)
    all_annotations_reference = []
    for row in reader:
        if len(row) == 1:
            continue
        else:
            all_annotations_reference.append(UniversalAnnotation.from_row(row))

    doc = nlp(data[text_name]["full_text"])
    add_annotation_to_tokens(doc, gold_standard_annos)
    lines = tokenise.tokenise(data[text_name]["full_text"])

    all_annotations_new = []
    for row in lines:
        if len(row) == 1:
            continue
        else:
            all_annotations_new.append(UniversalAnnotation.from_row(row.values()))

    strs_reference = [a.text for a in all_annotations_reference]
    strs_new = [a.text for a in all_annotations_new]

    matches = SequenceMatcher(None, strs_reference, strs_new)
    codes = matches.get_opcodes()
    annotations = get_reference_token_annotations(codes, doc, strs_reference)

    reader = csv.reader(open(reference_tsv), dialect=TSVDialect)
    i = 0
    with open(out_path, "w") as out_file:
        writer = csv.writer(out_file, dialect=TSVDialect)
        for row in reader:
            if len(row) == 1:
                writer.writerow(row)
            else:
                writer.writerow(row + list(annotations[i]))
                i += 1



if __name__ == "__main__":
    args = parser.parse_args()
    main(Path(args.zip_path), Path(args.tsv_input_path), args.text_name, args.tsv_output_path)
